#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Windows.Forms;
using HeuristicLab.Core;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;

namespace HeuristicLab.Parameters.Views {
  /// <summary>
  /// The visual representation of a <see cref="Parameter"/>.
  /// </summary>
  [View("ScopePathLookupParameter View")]
  [Content(typeof(ScopePathLookupParameter<>), true)]
  public partial class ScopePathLookupParameterView<T> : ParameterView where T : class, IItem {
    /// <summary>
    /// Gets or sets the variable to represent visually.
    /// </summary>
    /// <remarks>Uses property <see cref="ViewBase.Item"/> of base class <see cref="ViewBase"/>.
    /// No own data storage present.</remarks>
    public new ScopePathLookupParameter<T> Content {
      get { return (ScopePathLookupParameter<T>)base.Content; }
      set { base.Content = value; }
    }

    /// <summary>
    /// Initializes a new instance of <see cref="VariableView"/> with caption "Variable".
    /// </summary>
    public ScopePathLookupParameterView() {
      InitializeComponent();
    }

    /// <summary>
    /// Removes the eventhandlers from the underlying <see cref="IVariable"/>.
    /// </summary>
    /// <remarks>Calls <see cref="ViewBase.RemoveItemEvents"/> of base class <see cref="ViewBase"/>.</remarks>
    protected override void DeregisterContentEvents() {
      Content.ActualNameChanged -= new EventHandler(Content_ActualNameChanged);
      Content.PathChanged -= new EventHandler(Content_PathChanged);
      base.DeregisterContentEvents();
    }

    /// <summary>
    /// Adds eventhandlers to the underlying <see cref="IVariable"/>.
    /// </summary>
    /// <remarks>Calls <see cref="ViewBase.AddItemEvents"/> of base class <see cref="ViewBase"/>.</remarks>
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.ActualNameChanged += new EventHandler(Content_ActualNameChanged);
      Content.PathChanged += new EventHandler(Content_PathChanged);
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content == null) {
        actualNameTextBox.Text = string.Empty;
        pathTextBox.Text = string.Empty;
      } else {
        actualNameTextBox.Text = Content.ActualName;
        pathTextBox.Text = Content.Path;
      }
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      actualNameTextBox.Enabled = Content != null;
      actualNameTextBox.ReadOnly = ReadOnly;
      pathTextBox.Enabled = Content != null;
      pathTextBox.ReadOnly = ReadOnly;
    }

    protected virtual void Content_ActualNameChanged(object sender, EventArgs e) {
      if (InvokeRequired)
        Invoke(new EventHandler(Content_ActualNameChanged), sender, e);
      else
        actualNameTextBox.Text = Content.ActualName;
    }
    protected virtual void Content_PathChanged(object sender, EventArgs e) {
      if (InvokeRequired)
        Invoke(new EventHandler(Content_PathChanged), sender, e);
      else
        pathTextBox.Text = Content.Path;
    }

    protected virtual void actualNameTextBox_Validated(object sender, EventArgs e) {
      Content.ActualName = actualNameTextBox.Text;
    }
    protected virtual void pathTextBox_KeyDown(object sender, KeyEventArgs e) {
      if ((e.KeyCode == Keys.Enter) || (e.KeyCode == Keys.Return))
        pathLabel.Focus();  // set focus on label to validate data
      if (e.KeyCode == Keys.Escape) {
        pathTextBox.Text = Content.Path;
        pathLabel.Focus();  // set focus on label to validate data
      }
    }
    protected virtual void pathTextBox_Validating(object sender, System.ComponentModel.CancelEventArgs e) {
      if (!ScopePathLookupParameter<T>.ValidPathRegex.IsMatch(pathTextBox.Text)) {
        e.Cancel = true;
        errorProvider.SetError(pathTextBox, "Invalid Path");
        pathTextBox.SelectAll();
      }
    }
    protected virtual void pathTextBox_Validated(object sender, EventArgs e) {
      Content.Path = pathTextBox.Text;
      errorProvider.SetError(pathTextBox, string.Empty);
    }
  }
}
